#ifndef art_Framework_Core_FindOneP_h
#define art_Framework_Core_FindOneP_h
////////////////////////////////////////////////////////////////////////
// FindOneP
//
// A smart query object used as the main way of accessing associated
// objects in a one-to-one association.
//
// Given an Assns associating A with B (or B with A) (possibly with an
// associated data object D) and a sequence ACOLL of A objects to be
// found in the Assns, allow indexed access to the B and/or D objects
// associated with the A objects in ACOLL.
//
////////////////////////////////////
// Interface.
//////////
//
// For ease of understanding, the interface is presented here; reading
// the rest of the header is not for the faint of heart.
//
// Notes:
//
// * ProdB and Data are the only template arguments that must be
// specified when constructing a FindOneP. Any other items are deducible
// from arguments.
//
// * An attempt to create a FindOneP where one of the listed A objects
// actually has multiple B objects associated with it will result in an
// exception.
//
// * The FindOneP needs a source of objects of type A, an event and an
// input tag corresponding to the underlying association collection from
// which to create itself.
//
// * When constructed, the FindOneP will obtain and interrogate the
// correct Assns and provide access to the B (and/or D) object(s)
// associated with each supplied A object in the order in which the A
// objects were specified.
//
// * If the specified A does not have an associated B or D then the
// cet::maybe_ref will be invalid.
//
// * If the specified A has multiple associated Bs (or Ds) then an
// exception will be thrown: use FindMany instead.
//
// * If the required association collection has an extra data object D
// with each association then it *must* be specified as a template
// argument, even if it is not relevant to the current query.
//
// * *All* indexed accessors (at(), data(), get()) are bounds-checked.
//
// Useful typedefs.
//
// typedef ProdB assoc_t;
// typedef Data data_t;
// typedef typename std::vector<assoc_t const *>::size_type size_type; // FindOne
// typedef typename std::vector<art::Ptr<assoc_t> >::size_type size_type; // FindOneP
//
// Constructors.
//
// // From Handle to collection of A.
// FindOneP<ProdB>(Handle<ProdAColl> const &,
//                Event const &,
//                InputTag const &);
// FindOneP<ProdB, Data>(Handle<ProdAColl> const &,
//                      Event const &,
//                      InputTag const &);
//
// // From View<A>.
// FindOneP<ProdB>(View<ProdA> const &,
//                Event const &,
//                InputTag const &);
// FindOneP<ProdB, Data>(View<ProdA> const &,
//                      Event const &,
//                      InputTag const &);
//
// // From arbitrary sequence of Ptr<A>.
// FindOneP<ProdB>(PtrProdAColl const &,
//                Event const &,
//                InputTag const &);
// FindOneP<ProdB, Data>(PtrProdAColl const &,
//                      Event const &,
//                      InputTag const &);
//
// Modifiers.
//
// <NONE>.
//
// Accessors.
//
// size_type size() const;
// cet::maybe_ref<assoc_t const> at(size_type) const; // FindOne
// art::Ptr<assoc_t> const & at(size_type) const; // FindOneP
// cet::maybe_ref<data_t const> data(size_type) const;
// void get(size_type,
//          cet::maybe_ref<assoc_t const> &) const; // FindOne
// void get(size_type,
//          art::Ptr<assoc_t> &) const; // FindOneP
// void get(size_type,
//          cet::maybe_ref<assoc_t const> &,
//          cet::maybe_ref<data_t const> &)
//   const; // *Must* be used for FindOne<ProdB, Data>.
// void get(size_type,
//          art::Ptr<assoc_t> &,
//          cet::maybe_ref<data_t const> &)
//   const; // *Must* be used for FindOneP<ProdB, Data>.
//
// Comparison operations.
//
// bool operator == (FindOneP const & other) const;
//
////////////////////////////////////////////////////////////////////////
#include "art/Framework/Core/detail/IPRHelper.h"
#include "art/Framework/Core/detail/ProductIDProvider.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Utilities/InputTag.h"
#include "cetlib/maybe_ref.h"

#include <vector>

#define ART_IPR_BY_PTR

namespace art {
  // General template
  template <typename ProdB, typename DATA = void>
  class FindOneP;

  // Specialization.
  template <typename ProdB>
  class FindOneP<ProdB, void>;
}

////////////////////////////////////////////////////////////////////////
// Implementation of the specialization.
template <typename ProdB>
class art::FindOneP<ProdB, void> {
public:

  typedef ProdB assoc_t;

#ifdef ART_IPR_BY_PTR
  typedef std::vector<Ptr<assoc_t> > bColl_t;
#else
  typedef std::vector<ProdB const *> bColl_t;
#endif
  typedef typename bColl_t::value_type value_type;
  typedef typename bColl_t::size_type size_type;
  typedef typename bColl_t::difference_type difference_type;
  typedef typename bColl_t::const_reference const_reference;
  typedef typename bColl_t::reference reference;

  template <typename ProdAColl>
  FindOneP(Handle<ProdAColl> const & aCollection,
                       Event const & e,
                       InputTag const & tag);

  template <typename ProdA>
  FindOneP(View<ProdA> const & view,
                       Event const & e,
                       InputTag const & tag);

  template <typename PtrProdAColl>
  FindOneP(PtrProdAColl const & aPtrColl,
                       Event const & e,
                       InputTag const & tag);

  // Is this a valid query (did we find an Assns)?
  bool isValid() const;

  // Number of query results
  size_type size() const;

  // Associated item by index (bounds-checked).
#ifdef ART_IPR_BY_PTR
  Ptr<assoc_t> const & at(size_type i) const;
  void get(size_type i, Ptr<assoc_t> & item) const;
#else
  cet::maybe_ref<assoc_t const> at(size_type i) const;
  void get(size_type i, cet::maybe_ref<assoc_t const> & item) const;
#endif
  bool operator == (FindOneP<ProdB, void> const & other) const;

protected:
  FindOneP() : bCollection_() { }
  bColl_t & bCollection() { return bCollection_; }

private:
  void throwIfInvalid_() const;

  bColl_t bCollection_;
  std::shared_ptr<cet::exception const> storedException_;
};

template <typename ProdB, typename Data>
class art::FindOneP : private art::FindOneP<ProdB, void> {
private:
  typedef FindOneP<ProdB, void> base;
public:
  typedef std::vector<Data const *> dataColl_t;
  typedef typename base::value_type value_type;
  typedef typename base::size_type size_type;
  typedef typename base::difference_type difference_type;
  typedef typename base::const_reference const_reference;
  typedef typename base::reference reference;
  typedef typename base::assoc_t assoc_t;

  typedef Data const * data_const_pointer;
  typedef Data const & data_const_reference;
  typedef Data & data_reference;

  typedef Data data_t;

  template <typename ProdAColl>
  FindOneP(Handle<ProdAColl> const & aCollection,
                       Event const & e,
                       InputTag const & tag);

  template <typename ProdA>
  FindOneP(View<ProdA> const & view,
                       Event const & e,
                       InputTag const & tag);

  template <typename PtrProdAColl>
  FindOneP(PtrProdAColl const & aPtrColl,
                       Event const & e,
                       InputTag const & tag);

  using base::size;
  using base::isValid;
  using base::at;
  using base::get;

  // Association extra-data object by index (bounds-checked).
  cet::maybe_ref<Data const> data(size_type i) const;

  // Associated item and extra-data object by index (bounds-checked).
  void get(size_type i,
#ifdef ART_IPR_BY_PTR
           art::Ptr<assoc_t> & item,
#else
           cet::maybe_ref<assoc_t const> & item,
#endif
           cet::maybe_ref<Data const> & data) const;

  bool operator == (FindOneP<ProdB, Data> const & other) const;

private:
  dataColl_t dataCollection_;
};

////////////////////////////////////////////////////////////////////////
// Base class implementation.
template <typename ProdB>
template <typename ProdAColl>
art::FindOneP<ProdB, void>::
FindOneP(Handle<ProdAColl> const & aCollection,
                     Event const & e,
                     InputTag const & tag)
  :
  bCollection_(),
  storedException_()
{
  typedef typename ProdAColl::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, void, void> finder(e, tag);
  storedException_ =
    finder(*aCollection, detail::ConstantProductIDProvider(aCollection.id()), bCollection_);
}

template <typename ProdB>
template <typename ProdA>
art::FindOneP<ProdB, void>::
FindOneP(View<ProdA> const & view,
                     Event const & e,
                     InputTag const & tag)
  :
  bCollection_(),
  storedException_()
{
  typename View<ProdA>::collection_type const & aCollection(view.vals());
  detail::IPRHelper<ProdA, ProdB, void, void> finder(e, tag);
  storedException_ =
    finder(aCollection, detail::ConstantProductIDProvider(view.id()), bCollection_);
}

template <typename ProdB>
template <typename PtrProdAColl>
art::FindOneP<ProdB, void>::
FindOneP(PtrProdAColl const & aPtrColl,
                     Event const & e,
                     InputTag const & tag)
  :
  bCollection_(),
  storedException_()
{
  typedef typename PtrProdAColl::value_type::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, void, void> finder(e, tag);
  storedException_ =
    finder(aPtrColl, detail::ProductIDProvider(), bCollection_);
}

template <typename ProdB>
inline
typename art::FindOneP<ProdB, void>::size_type
art::FindOneP<ProdB, void>::size() const
{
  throwIfInvalid_();
  return bCollection_.size();
}

template <typename ProdB>
inline
bool
art::FindOneP<ProdB, void>::isValid() const
{
  return (storedException_.get() == nullptr);
}

template <typename ProdB>
inline
#ifdef ART_IPR_BY_PTR
art::Ptr<typename art::FindOneP<ProdB, void>::assoc_t> const &
#else
cet::maybe_ref < typename art::FindOneP<ProdB, void>::assoc_t const >
#endif
art::FindOneP<ProdB, void>::at(size_type i) const
{
  throwIfInvalid_();
#ifdef ART_IPR_BY_PTR
  return bCollection_.at(i);
#else
  return cet::maybe_ref<assoc_t const>(*bCollection_.at(i));
#endif
}

template <typename ProdB>
inline
void
art::FindOneP<ProdB, void>::
get(size_type i,
#ifdef ART_IPR_BY_PTR
    Ptr<assoc_t> & item
#else
    cet::maybe_ref<assoc_t const> & item
#endif
   ) const
{
  throwIfInvalid_();
#ifdef ART_IPR_BY_PTR
  item = bCollection_.at(i);
#else
  item.reseat(*bCollection_.at(i));
#endif
}

template <typename ProdB>
inline
bool
art::FindOneP<ProdB, void>::operator == (FindOneP<ProdB, void> const & other) const
{
  throwIfInvalid_();
  return bCollection_ == other.bCollection_;
}

template <typename ProdB>
inline
void
art::FindOneP<ProdB, void>::throwIfInvalid_() const
{
  if (!isValid()) {
    throw Exception(errors::LogicError, "Invalid FindOneP", *storedException_)
      << "Attempt to use a FindOneP where the underlying art::Assns product was not found.";
  }
}

////////////////////////////////////////////////////////////////////////
// Derived class implementation.
template <typename ProdB, typename Data>
template <typename ProdAColl>
art::FindOneP<ProdB, Data>::FindOneP(Handle<ProdAColl> const & aCollection,
    Event const & e,
    InputTag const & tag)
  :
  base(),
  dataCollection_()
{
  typedef typename ProdAColl::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t> finder(e, tag);
  finder(*aCollection, detail::ConstantProductIDProvider(aCollection.id()), base::bCollection(), dataCollection_);
}

template <typename ProdB, typename Data>
template <typename ProdA>
art::FindOneP<ProdB, Data>::FindOneP(View<ProdA> const & view,
    Event const & e,
    InputTag const & tag)
  :
  base(),
  dataCollection_()
{
  typename View<ProdA>::collection_type const & aCollection(view.vals());
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t> finder(e, tag);
  finder(aCollection, detail::ConstantProductIDProvider(view.id()), base::bCollection(), dataCollection_);
}

template <typename ProdB, typename Data>
template <typename PtrProdAColl>
art::FindOneP<ProdB, Data>::FindOneP(PtrProdAColl const & aPtrColl,
    Event const & e,
    InputTag const & tag)
  :
  base(),
  dataCollection_()
{
  typedef typename PtrProdAColl::value_type::value_type ProdA;
  detail::IPRHelper<ProdA, ProdB, Data, dataColl_t> finder(e, tag);
  finder(aPtrColl, detail::ProductIDProvider(), base::bCollection(), dataCollection_);
}

template <typename ProdB, typename Data>
inline
cet::maybe_ref<Data const>
art::FindOneP<ProdB, Data>::data(size_type i) const
{
  return cet::maybe_ref<Data const>(*dataCollection_.at(i));
}

template <typename ProdB, typename Data>
inline
void
art::FindOneP<ProdB, Data>::
get(size_type i,
#ifdef ART_IPR_BY_PTR
    Ptr<assoc_t> & item,
#else
    cet::maybe_ref<assoc_t const> & item,
#endif
    cet::maybe_ref<Data const> & data) const
{
  base::get(i, item);
  data.reseat(*dataCollection_.at(i));
}

template <typename ProdB, typename Data>
inline
bool
art::FindOneP<ProdB, Data>::operator == (FindOneP<ProdB, Data> const & other) const
{
  return dataCollection_ == other.dataCollection_ &&
         this->base::operator==(other);
}

#undef ART_IPR_BY_PTR

#endif /* art_Framework_Core_FindOneP_h */

// Local Variables:
// mode: c++
// End:
